/*
 * Author:  Bastien Tauran
 * Date:    4.11.2016
 * Edited:	20.12.2016
 *
 * Scheduling policies for A-HARQ module in ns-2
 *
 * This file implements the global behavior of A-HARQ queue
 */

#ifndef tango_scheduler_h
#define tango_scheduler_h

#include <string.h>
#include <iostream>
#include <vector>
#include <map>
#include "queue.h"
#include "config.h"
#include "TANGOQueue.h"
#include "tclcl.h"

class HARQQueue;
class HARQT;
class HARQR;
class TANGOQueue;

class HARQHandler : public Handler {
public:
    HARQHandler(HARQQueue *queue) : queue_(queue) {};
    void handle(Event*);
private: 
    HARQQueue *queue_;
};

/**
* Global queue of TANGO scheduler. This Queue contains several TANGOQueue

	HARQQueue (scheduler)
	+------------------+
	|                  |
	| +--------------+ |
	| |  TANGOQueue  | |
	| +--------------+ |
	|                  |
	| +--------------+ |
	| |  TANGOQueue  | |
	| +--------------+ |
	|                  |
	| +--------------+ |
	| |  TANGOQueue  | |
	| +--------------+ |
	|                  |
	+------------------+

*/
class HARQQueue : public Queue {
	public:
		HARQQueue();									// Constructor
		
		bool blocked_HARQ_;
		HARQHandler harqh_;

	protected:
		void reset();									// Resets the queue
		bool isNewAddress(int);							// Checks if the address already exists
		void enque(Packet*);							// Enque the packet in the correct queue
		Packet* deque();								// Remove the packet in the queue selected depending on the scheduler rule
		virtual TANGOQueue* getAddressDeque();			// Returns the queue chosen
		void print_trace_queue(char,int,int,int,int);	// Prints values in a trace file
		int command(int, const char*const*);			// Link to otcl
		void results();									// Prints the statistics

		vector<TANGOQueue*> qs_;						// List of queues
		PacketQueue* queueFEC_;							// To store FEC packets
		int debug_;										// To print debugging information
		BaseTrace *pt_queue_;							// Used for printing debugging
		HARQT* harqt_;									// HARQ Sender
		HARQR* harqr_;									// HARQ Receiver
		int dropped_;									// Counts the number of packets dropped by the queue
		int queue_size_max_;							// Queue size in packets
		double time_max_queue_;							// Timeout for PFDrop
};

/* Round Robin scheduling policy */
class HARQQueueRR : public HARQQueue {
	public:
		HARQQueueRR()
		{
			cout << "New RR Queue" << endl;
			bind_bool("debug_", &debug_);
			index_current_ = 0;
			queueFEC_ = new PacketQueue();
		}
	protected:
		TANGOQueue* getAddressDeque();					// Returns the queue chosen with Round Robin
		int index_current_;								// Index of current queue chosen
};

/* MLWDF scheduling policy */
class HARQQueueMLWDF : public HARQQueue {
	public:
		HARQQueueMLWDF()
		{
			cout << "New MLWDF Queue" << endl;
			bind_bool("debug_", &debug_);
			queueFEC_ = new PacketQueue();
		}
	protected:
		TANGOQueue* getAddressDeque();					// Returns the queue chosen with M-LWDF
};

/* Proportional Fairness scheduling policy */
class HARQQueuePF : public HARQQueue {
	public:
		HARQQueuePF()
		{
			cout << "New PF Queue" << endl;
			bind_bool("debug_", &debug_);
			queueFEC_ = new PacketQueue();
		}
	protected:
		TANGOQueue* getAddressDeque();					// Returns the queue chosen with PF
};

/* CoDeS scheduling policy */
class HARQQueuePFDrop : public HARQQueue {
	public:
		HARQQueuePFDrop()
		{
			cout << "New PF Drop Queue" << endl;
			bind_bool("debug_", &debug_);
			queueFEC_ = new PacketQueue();
		}
	protected:
		TANGOQueue* getAddressDeque();					// Returns the queue chosen with PFDrop
};

/* FIFO queueing policy */
class HARQQueueFIFO : public HARQQueue {
	public:
		HARQQueueFIFO()
		{
			cout << "New No Queue" << endl;
			bind_bool("debug_", &debug_);
			queueFEC_ = new PacketQueue();
		}
	protected:
		TANGOQueue* getAddressDeque();					// Returns the queue
		void enque(Packet*);							// Enque the packet
};

#endif
