/*
 * Author:  Bastien Tauran
 * Date:    4.11.2016
 * Edited:	20.12.2016
 *
 * Adaptive-HARQ module
 *
 * Adaptation of A-HARQ to NS-2
 */

#ifndef ADAPTIVE_HARQ_H
#define ADAPTIVE_HARQ_H

#include <iostream>
#include <sstream>
#include <ctime>
#include <string>
#include "tclcl.h"
#include "agent.h"
#include "flags.h"
#include "ip.h"
#include "tcp.h"
#include "random.h"
#include "packet.h"
#include "basetrace.h"
#include "queue.h"
#include "node.h"
#include "TANGOScheduler.h"

/* Struct to store data linked with one packet */
typedef struct pkt_recv
{
	bool valid;
	Packet* p;
	Handler* h;
	int seqno;
	int id;
	int number_ack_sent;
	double time_in;
} pkt_recv;

/* Struct to store data linked with one FID */
typedef struct expected_struct
{
	int id;
	int expected;
	double last_time;
	bool valid;
} expected_struct;

/* Header for HARQ */
typedef struct hdr_harq
{
	int nb_bits_;
	int for_id_;
	int size_;
	int real_size_;
	double accumulated_IM_;
	int LMS_offset_;
	static int offset_;
	int is_fec_;
	
	inline static int& offset() { return offset_; }
	inline static hdr_harq* access(const Packet* p) {
		return (hdr_harq*) p->access(offset_);
	}
	int& nb_bits() { return (nb_bits_); }
	int& for_id() { return (for_id_); }
	int& size() { return (size_); }
	int& real_size() { return (real_size_); }
	double& accumulated_IM() { return (accumulated_IM_); }
	int& LMS_offset() { return (LMS_offset_); }
} hdr_harq;

/* HARQ packet header */
static class AckHarqHeaderClass : public PacketHeaderClass {
public:
	AckHarqHeaderClass() : PacketHeaderClass("PacketHeader/ACK_HARQ", sizeof(hdr_harq)) {
	 	bind_offset(&hdr_harq::offset_);
	}
} class_hdr_harq;

/* Mother class */
class HARQ : public Agent { 
	public:
		HARQ(packet_t);
	protected:
		double delay_;								// Transmission time between sender and receiver
		int Es_N0_;									// Attenuation
		int n_;										// Number of bits per symbol
		
		static double error_prob[7][4];				// Probability of an error at each retransmission
		static int nb_symbols_to_send[7][4];		// Number of symbols to send at each retransmission
};

class HARQR;
class SatelliteScheduler;

/* Sender side of HARQ */
class HARQT : public HARQ { 
	public:
		HARQT();
		void recv(Packet*, Handler*);						// When a packet is received
		void setHARQR(HARQR*);								// Set the receiver
		void setDelay(double);								// Set the delay
		double getDelay();									// Get the delay
		void setEsN0(int);									// Set the attenuation
		int command(int, const char*const*);
	protected:
		void sendFEC(int, int, Packet*, double,Handler*);	// Send FEC bits to the receiver
		
		HARQR* harqr_;										// Pointer to the receiver
		SatelliteScheduler* scheduler_;						// Scheduler for packets
		int debug_;											// To debug
		Queue* queue_;										// TANGOQueue chosen
};

/* Receiver side of HARQ */
class HARQR : public HARQ { 
	public:
		HARQR();
		void recv(Packet*, Handler*);										// When a packet is received
		int command(int, const char*const*);								// To add commands for OTCL
	protected:
		void ack(Packet*, int, int, double);								// Send an ACK to the sender
		int getFirstFree();													// Return the first free space in the buffer, -1 otherwise
		int already(Packet*);												// Test if a packet is already in a buffer and returns its index, -1 otherwise
		int addToBuffer(Packet*, Handler*, int);							// Add a packet to the buffer
		void removeFromBuffer(int);											// Remove a packet from the buffer
		void forward_pkt(Packet*,Handler*,int,int);							// Forward a packet when it is decoded or put it in buffer_out if reordering is needed
		bool isFavor(int, int);												// Return if the packet is favored, offset is 0 if the packet is in the buffer, 1 otherwise
		int isFull();														// Test if buffer_out is full, and return the index of the first empty space, -1 otherwise
		bool enque(pkt_recv);												// Put a packet if buffer_out
		bool deque(pkt_recv*,int,int);										// Remove a packet from buffer_out
		void print_trace_reconstruction(Packet*, int, char, int, int, int);	// Print to the reconstruction trace file
		void print_trace_reordering(Packet*, char, int, int);				// Print to the reordering trace file
		void change_buffer_out_size(int);									// Change the size of buffer_out and reinitialize it
		void change_buffer_fid_size(int);									// Change the size of buffer_fid and reinitialize it
		virtual double getError(int, double, int);							// See HARQStatic
		virtual int computeBitsToSend(int);									// See HARQStatic
		void results();														// Print the results
		
		int buffer_reordering_size_;				// Size of reordering buffer
		int buffer_fid_size_;						// Size of expected_list_
		pkt_recv* buffer_reconstruction_;			// Buffer to store packets not decoded yet
		pkt_recv* buffer_reordering_;				// Buffer to store packets not ordered
		expected_struct* expected_list_;			// List to store the seqno expected for each FID
		int buffer_reordering_filling_;				// Number of packets in buffer_reordering_
		HARQT* harqt_;								// Pointer to the sender
		int debug_;									// To debug
		int trace_reconstruction_;					// Print HARQ to a trace file
		int trace_reordering_;						// Print reordering to a trace file
		int reordering_;							// Using reordering or not
		int favor_;									// Using FavorQueue policy or not
		double timeout_;							// Value of timeout if there is one
		int dropped_;								// Number of dropped packets by HARQ
		int dropped_reordering_buffer_;				// Number of packets dropped by the reordering buffer
		int* forwarded_;							// Number of packets retransmitted for each number of retransmission
		BaseTrace *pt_reconstruction_;				// Support for pkt reconstruction tracing
		BaseTrace *pt_reordering_;					// Support for pkt reordering tracing
};

#endif
